/****************************************************************************
 Title	:   HD44780U LCD2 library
 Author:    Peter Fleury <pfleury@gmx.ch>  http://jump.to/fleury
 File:	    $Id: lcd2.c,v 1.14.2.1 2006/01/29 12:16:41 peter Exp $
 Software:  AVR-GCC 3.3 
 Target:    any AVR device, memory mapped mode only for AT90S4414/8515/Mega

 DESCRIPTION
       Basic routines for interfacing a HD44780U-based text lcd2 display

       Originally based on Volker Oth's lcd2 library,
       changed lcd2_init(), added additional constants for lcd2_command(),
       added 4-bit I/O mode, improved and optimized code.

       Library can be operated in memory mapped mode (LCD2_IO_MODE=0) or in 
       4-bit IO port mode (LCD2_IO_MODE=1). 8-bit IO port mode not supported.
       
       Memory mapped mode compatible with Kanda STK200, but supports also
       generation of R/W signal through A8 address line.

 USAGE
       See the C include lcd2.h file for a description of each function
       
*****************************************************************************/
#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "lcd_02.h"



/* 
** constants/macros 
*/
#define DDR(x) (*(&x - 1))      /* address of data direction register of port x */
#if defined(__AVR_ATmega64__) || defined(__AVR_ATmega128__)
    /* on ATmega64/128 PINF is on port 0x00 and not 0x60 */
    #define PIN(x) ( &PORTF==&(x) ? _SFR_IO8(0x00) : (*(&x - 2)) )
#else
	#define PIN(x) (*(&x - 2))    /* address of input register of port x          */
#endif


#if LCD2_IO_MODE
#define lcd2_e_delay()   __asm__ __volatile__( "rjmp 1f\n 1:" );
#define lcd2_e_high()    LCD2_E_PORT  |=  _BV(LCD2_E_PIN);
#define lcd2_e_low()     LCD2_E_PORT  &= ~_BV(LCD2_E_PIN);
#define lcd2_e_toggle()  toggle_e()
#define lcd2_rw_high()   LCD2_RW_PORT |=  _BV(LCD2_RW_PIN)
#define lcd2_rw_low()    LCD2_RW_PORT &= ~_BV(LCD2_RW_PIN)
#define lcd2_rs_high()   LCD2_RS_PORT |=  _BV(LCD2_RS_PIN)
#define lcd2_rs_low()    LCD2_RS_PORT &= ~_BV(LCD2_RS_PIN)
#endif

#if LCD2_IO_MODE
#if LCD2_LINES==1
#define LCD2_FUNCTION_DEFAULT    LCD2_FUNCTION_4BIT_1LINE 
#else
#define LCD2_FUNCTION_DEFAULT    LCD2_FUNCTION_4BIT_2LINES 
#endif
#else
#if LCD2_LINES==1
#define LCD2_FUNCTION_DEFAULT    LCD2_FUNCTION_8BIT_1LINE
#else
#define LCD2_FUNCTION_DEFAULT    LCD2_FUNCTION_8BIT_2LINES
#endif
#endif

#if LCD2_CONTROLLER_KS0073
#if LCD2_LINES==4

#define KS0073_EXTENDED_FUNCTION_REGISTER_ON  0x24   /* |0|010|0100 4-bit mode extension-bit RE = 1 */
#define KS0073_EXTENDED_FUNCTION_REGISTER_OFF 0x20   /* |0|000|1001 4 lines mode */
#define KS0073_4LINES_MODE                    0x09   /* |0|001|0000 4-bit mode, extension-bit RE = 0 */

#endif
#endif

/* 
** function prototypes 
*/
#if LCD2_IO_MODE
static void toggle_e(void);
#endif

/*
** local functions
*/



/*************************************************************************
 delay loop for small accurate delays: 16-bit counter, 4 cycles/loop
*************************************************************************/
static inline void _delayFourCycles(unsigned int __count)
{
    if ( __count == 0 )    
        __asm__ __volatile__( "rjmp 1f\n 1:" );    // 2 cycles
    else
        __asm__ __volatile__ (
    	    "1: sbiw %0,1" "\n\t"                  
    	    "brne 1b"                              // 4 cycles/loop
    	    : "=w" (__count)
    	    : "0" (__count)
    	   );
}


/************************************************************************* 
delay for a minimum of <us> microseconds
the number of loops is calculated at compile-time from MCU clock frequency
*************************************************************************/
#define delay(us)  _delayFourCycles( ( ( 1*(XTAL/4000) )*us)/1000 )


#if LCD2_IO_MODE
/* toggle Enable Pin to initiate write */
static void toggle_e(void)
{
    lcd2_e_high();
    lcd2_e_delay();
    lcd2_e_low();
}
#endif


/*************************************************************************
Low-level function to write byte to LCD2 controller
Input:    data   byte to write to LCD2
          rs     1: write data    
                 0: write instruction
Returns:  none
*************************************************************************/
#if LCD2_IO_MODE
static void lcd2_write(uint8_t data,uint8_t rs) 
{
    unsigned char dataBits ;


    if (rs) {   /* write data        (RS=1, RW=0) */
       lcd2_rs_high();
    } else {    /* write instruction (RS=0, RW=0) */
       lcd2_rs_low();
    }
    lcd2_rw_low();

    if ( ( &LCD2_DATA0_PORT == &LCD2_DATA1_PORT) && ( &LCD2_DATA1_PORT == &LCD2_DATA2_PORT ) && ( &LCD2_DATA2_PORT == &LCD2_DATA3_PORT )
      && (LCD2_DATA0_PIN == 0) && (LCD2_DATA1_PIN == 1) && (LCD2_DATA2_PIN == 2) && (LCD2_DATA3_PIN == 3) )
    {
        /* configure data pins as output */
        DDR(LCD2_DATA0_PORT) |= 0x0F;

        /* output high nibble first */
        dataBits = LCD2_DATA0_PORT & 0xF0;
        LCD2_DATA0_PORT = dataBits |((data>>4)&0x0F);
        lcd2_e_toggle();

        /* output low nibble */
        LCD2_DATA0_PORT = dataBits | (data&0x0F);
        lcd2_e_toggle();

        /* all data pins high (inactive) */
        LCD2_DATA0_PORT = dataBits | 0x0F;
    }
    else
    {
        /* configure data pins as output */
        DDR(LCD2_DATA0_PORT) |= _BV(LCD2_DATA0_PIN);
        DDR(LCD2_DATA1_PORT) |= _BV(LCD2_DATA1_PIN);
        DDR(LCD2_DATA2_PORT) |= _BV(LCD2_DATA2_PIN);
        DDR(LCD2_DATA3_PORT) |= _BV(LCD2_DATA3_PIN);
        
        /* output high nibble first */
        LCD2_DATA3_PORT &= ~_BV(LCD2_DATA3_PIN);
        LCD2_DATA2_PORT &= ~_BV(LCD2_DATA2_PIN);
        LCD2_DATA1_PORT &= ~_BV(LCD2_DATA1_PIN);
        LCD2_DATA0_PORT &= ~_BV(LCD2_DATA0_PIN);
    	if(data & 0x80) LCD2_DATA3_PORT |= _BV(LCD2_DATA3_PIN);
    	if(data & 0x40) LCD2_DATA2_PORT |= _BV(LCD2_DATA2_PIN);
    	if(data & 0x20) LCD2_DATA1_PORT |= _BV(LCD2_DATA1_PIN);
    	if(data & 0x10) LCD2_DATA0_PORT |= _BV(LCD2_DATA0_PIN);   
        lcd2_e_toggle();
        
        /* output low nibble */
        LCD2_DATA3_PORT &= ~_BV(LCD2_DATA3_PIN);
        LCD2_DATA2_PORT &= ~_BV(LCD2_DATA2_PIN);
        LCD2_DATA1_PORT &= ~_BV(LCD2_DATA1_PIN);
        LCD2_DATA0_PORT &= ~_BV(LCD2_DATA0_PIN);
    	if(data & 0x08) LCD2_DATA3_PORT |= _BV(LCD2_DATA3_PIN);
    	if(data & 0x04) LCD2_DATA2_PORT |= _BV(LCD2_DATA2_PIN);
    	if(data & 0x02) LCD2_DATA1_PORT |= _BV(LCD2_DATA1_PIN);
    	if(data & 0x01) LCD2_DATA0_PORT |= _BV(LCD2_DATA0_PIN);
        lcd2_e_toggle();        
        
        /* all data pins high (inactive) */
        LCD2_DATA0_PORT |= _BV(LCD2_DATA0_PIN);
        LCD2_DATA1_PORT |= _BV(LCD2_DATA1_PIN);
        LCD2_DATA2_PORT |= _BV(LCD2_DATA2_PIN);
        LCD2_DATA3_PORT |= _BV(LCD2_DATA3_PIN);
    }
}
#else
#define lcd2_write(d,rs) if (rs) *(volatile uint8_t*)(LCD2_IO_DATA) = d; else *(volatile uint8_t*)(LCD2_IO_FUNCTION) = d;
/* rs==0 -> write instruction to LCD2_IO_FUNCTION */
/* rs==1 -> write data to LCD2_IO_DATA */
#endif


/*************************************************************************
Low-level function to read byte from LCD2 controller
Input:    rs     1: read data    
                 0: read busy flag / address counter
Returns:  byte read from LCD2 controller
*************************************************************************/
#if LCD2_IO_MODE
static uint8_t lcd2_read(uint8_t rs) 
{
    uint8_t data;
    
    
    if (rs)
        lcd2_rs_high();                       /* RS=1: read data      */
    else
        lcd2_rs_low();                        /* RS=0: read busy flag */
    lcd2_rw_high();                           /* RW=1  read mode      */
    
    if ( ( &LCD2_DATA0_PORT == &LCD2_DATA1_PORT) && ( &LCD2_DATA1_PORT == &LCD2_DATA2_PORT ) && ( &LCD2_DATA2_PORT == &LCD2_DATA3_PORT )
      && ( LCD2_DATA0_PIN == 0 )&& (LCD2_DATA1_PIN == 1) && (LCD2_DATA2_PIN == 2) && (LCD2_DATA3_PIN == 3) )
    {
        DDR(LCD2_DATA0_PORT) &= 0xF0;         /* configure data pins as input */
        
        lcd2_e_high();
        lcd2_e_delay();        
        data = PIN(LCD2_DATA0_PORT) << 4;     /* read high nibble first */
        lcd2_e_low();
        
        lcd2_e_delay();                       /* Enable 500ns low       */
        
        lcd2_e_high();
        lcd2_e_delay();
        data |= PIN(LCD2_DATA0_PORT)&0x0F;    /* read low nibble        */
        lcd2_e_low();
    }
    else
    {
        /* configure data pins as input */
        DDR(LCD2_DATA0_PORT) &= ~_BV(LCD2_DATA0_PIN);
        DDR(LCD2_DATA1_PORT) &= ~_BV(LCD2_DATA1_PIN);
        DDR(LCD2_DATA2_PORT) &= ~_BV(LCD2_DATA2_PIN);
        DDR(LCD2_DATA3_PORT) &= ~_BV(LCD2_DATA3_PIN);
                
        /* read high nibble first */
        lcd2_e_high();
        lcd2_e_delay();        
        data = 0;
        if ( PIN(LCD2_DATA0_PORT) & _BV(LCD2_DATA0_PIN) ) data |= 0x10;
        if ( PIN(LCD2_DATA1_PORT) & _BV(LCD2_DATA1_PIN) ) data |= 0x20;
        if ( PIN(LCD2_DATA2_PORT) & _BV(LCD2_DATA2_PIN) ) data |= 0x40;
        if ( PIN(LCD2_DATA3_PORT) & _BV(LCD2_DATA3_PIN) ) data |= 0x80;
        lcd2_e_low();

        lcd2_e_delay();                       /* Enable 500ns low       */
    
        /* read low nibble */    
        lcd2_e_high();
        lcd2_e_delay();
        if ( PIN(LCD2_DATA0_PORT) & _BV(LCD2_DATA0_PIN) ) data |= 0x01;
        if ( PIN(LCD2_DATA1_PORT) & _BV(LCD2_DATA1_PIN) ) data |= 0x02;
        if ( PIN(LCD2_DATA2_PORT) & _BV(LCD2_DATA2_PIN) ) data |= 0x04;
        if ( PIN(LCD2_DATA3_PORT) & _BV(LCD2_DATA3_PIN) ) data |= 0x08;        
        lcd2_e_low();
    }
    return data;
}
#else
#define lcd2_read(rs) (rs) ? *(volatile uint8_t*)(LCD2_IO_DATA+LCD2_IO_READ) : *(volatile uint8_t*)(LCD2_IO_FUNCTION+LCD2_IO_READ)
/* rs==0 -> read instruction from LCD2_IO_FUNCTION */
/* rs==1 -> read data from LCD2_IO_DATA */
#endif


/*************************************************************************
loops while lcd2 is busy, returns address counter
*************************************************************************/
static uint8_t lcd2_waitbusy(void)

{
    register uint8_t c;
    
    /* wait until busy flag is cleared */
    while ( (c=lcd2_read(0)) & (1<<LCD2_BUSY)) {}
    
    /* the address counter is updated 4us after the busy flag is cleared */
    delay(2);

    /* now read the address counter */
    return (lcd2_read(0));  // return address counter
    
}/* lcd2_waitbusy */


/*************************************************************************
Move cursor to the start of next line or to the first line if the cursor 
is already on the last line.
*************************************************************************/
static inline void lcd2_newline(uint8_t pos)
{
    register uint8_t addressCounter;


#if LCD2_LINES==1
    addressCounter = 0;
#endif
#if LCD2_LINES==2
    if ( pos < (LCD2_START_LINE2) )
        addressCounter = LCD2_START_LINE2;
    else
        addressCounter = LCD2_START_LINE1;
#endif
#if LCD2_LINES==4
#if KS0073_4LINES_MODE
    if ( pos < LCD2_START_LINE2 )
        addressCounter = LCD2_START_LINE2;
    else if ( (pos >= LCD2_START_LINE2) && (pos < LCD2_START_LINE3) )
        addressCounter = LCD2_START_LINE3;
    else if ( (pos >= LCD2_START_LINE3) && (pos < LCD2_START_LINE4) )
        addressCounter = LCD2_START_LINE4;
    else 
        addressCounter = LCD2_START_LINE1;
#else
    if ( pos < LCD2_START_LINE3 )
        addressCounter = LCD2_START_LINE2;
    else if ( (pos >= LCD2_START_LINE2) && (pos < LCD2_START_LINE4) )
        addressCounter = LCD2_START_LINE3;
    else if ( (pos >= LCD2_START_LINE3) && (pos < LCD2_START_LINE2) )
        addressCounter = LCD2_START_LINE4;
    else 
        addressCounter = LCD2_START_LINE1;
#endif
#endif
    lcd2_command((1<<LCD2_DDRAM)+addressCounter);

}/* lcd2_newline */


/*
** PUBLIC FUNCTIONS 
*/

/*************************************************************************
Send LCD2 controller instruction command
Input:   instruction to send to LCD2 controller, see HD44780 data sheet
Returns: none
*************************************************************************/
void lcd2_command(uint8_t cmd)
{
    lcd2_waitbusy();
    lcd2_write(cmd,0);
}


/*************************************************************************
Send data byte to LCD2 controller 
Input:   data to send to LCD2 controller, see HD44780 data sheet
Returns: none
*************************************************************************/
void lcd2_data(uint8_t data)
{
    lcd2_waitbusy();
    lcd2_write(data,1);
}



/*************************************************************************
Set cursor to specified position
Input:    x  horizontal position  (0: left most position)
          y  vertical position    (0: first line)
Returns:  none
*************************************************************************/
void lcd2_gotoxy(uint8_t x, uint8_t y)
{
#if LCD2_LINES==1
    lcd2_command((1<<LCD2_DDRAM)+LCD2_START_LINE1+x);
#endif
#if LCD2_LINES==2
    if ( y==0 ) 
        lcd2_command((1<<LCD2_DDRAM)+LCD2_START_LINE1+x);
    else
        lcd2_command((1<<LCD2_DDRAM)+LCD2_START_LINE2+x);
#endif
#if LCD2_LINES==4
    if ( y==0 )
        lcd2_command((1<<LCD2_DDRAM)+LCD2_START_LINE1+x);
    else if ( y==1)
        lcd2_command((1<<LCD2_DDRAM)+LCD2_START_LINE2+x);
    else if ( y==2)
        lcd2_command((1<<LCD2_DDRAM)+LCD2_START_LINE3+x);
    else /* y==3 */
        lcd2_command((1<<LCD2_DDRAM)+LCD2_START_LINE4+x);
#endif

}/* lcd2_gotoxy */


/*************************************************************************
*************************************************************************/
int lcd2_getxy(void)
{
    return lcd2_waitbusy();
}


/*************************************************************************
Clear display and set cursor to home position
*************************************************************************/
void lcd2_clrscr(void)
{
    lcd2_command(1<<LCD2_CLR);
}


/*************************************************************************
Set cursor to home position
*************************************************************************/
void lcd2_home(void)
{
    lcd2_command(1<<LCD2_HOME);
}


/*************************************************************************
Display character at current cursor position 
Input:    character to be displayed                                       
Returns:  none
*************************************************************************/
void lcd2_putc(char c)
{
    uint8_t pos;


    pos = lcd2_waitbusy();   // read busy-flag and address counter
    if (c=='\n')
    {
        lcd2_newline(pos);
    }
    else
    {
#if LCD2_WRAP_LINES==1
#if LCD2_LINES==1
        if ( pos == LCD2_START_LINE1+LCD2_DISP_LENGTH ) {
            lcd2_write((1<<LCD2_DDRAM)+LCD2_START_LINE1,0);
        }
#elif LCD2_LINES==2
        if ( pos == LCD2_START_LINE1+LCD2_DISP_LENGTH ) {
            lcd2_write((1<<LCD2_DDRAM)+LCD2_START_LINE2,0);    
        }else if ( pos == LCD2_START_LINE2+LCD2_DISP_LENGTH ){
            lcd2_write((1<<LCD2_DDRAM)+LCD2_START_LINE1,0);
        }
#elif LCD2_LINES==4
        if ( pos == LCD2_START_LINE1+LCD2_DISP_LENGTH ) {
            lcd2_write((1<<LCD2_DDRAM)+LCD2_START_LINE2,0);    
        }else if ( pos == LCD2_START_LINE2+LCD2_DISP_LENGTH ) {
            lcd2_write((1<<LCD2_DDRAM)+LCD2_START_LINE3,0);
        }else if ( pos == LCD2_START_LINE3+LCD2_DISP_LENGTH ) {
            lcd2_write((1<<LCD2_DDRAM)+LCD2_START_LINE4,0);
        }else if ( pos == LCD2_START_LINE4+LCD2_DISP_LENGTH ) {
            lcd2_write((1<<LCD2_DDRAM)+LCD2_START_LINE1,0);
        }
#endif
        lcd2_waitbusy();
#endif
        lcd2_write(c, 1);
    }

}/* lcd2_putc */


/*************************************************************************
Display string without auto linefeed 
Input:    string to be displayed
Returns:  none
*************************************************************************/
void lcd2_puts(const char *s)
/* print string on lcd2 (no auto linefeed) */
{
    register char c;

    while ( (c = *s++) ) {
        lcd2_putc(c);
    }

}/* lcd2_puts */


/*************************************************************************
Display string from program memory without auto linefeed 
Input:     string from program memory be be displayed                                        
Returns:   none
*************************************************************************/
void lcd2_puts_p(const char *progmem_s)
/* print string from program memory on lcd2 (no auto linefeed) */
{
    register char c;

    while ( (c = pgm_read_byte(progmem_s++)) ) {
        lcd2_putc(c);
    }

}/* lcd2_puts_p */


/*************************************************************************
Initialize display and select type of cursor 
Input:    dispAttr LCD2_DISP_OFF            display off
                   LCD2_DISP_ON             display on, cursor off
                   LCD2_DISP_ON_CURSOR      display on, cursor on
                   LCD2_DISP_CURSOR_BLINK   display on, cursor on flashing
Returns:  none
*************************************************************************/
void lcd2_init(uint8_t dispAttr)
{
#if LCD2_IO_MODE
    /*
     *  Initialize LCD2 to 4 bit I/O mode
     */
     
    if ( ( &LCD2_DATA0_PORT == &LCD2_DATA1_PORT) && ( &LCD2_DATA1_PORT == &LCD2_DATA2_PORT ) && ( &LCD2_DATA2_PORT == &LCD2_DATA3_PORT )
      && ( &LCD2_RS_PORT == &LCD2_DATA0_PORT) && ( &LCD2_RW_PORT == &LCD2_DATA0_PORT) && (&LCD2_E_PORT == &LCD2_DATA0_PORT)
      && (LCD2_DATA0_PIN == 0 ) && (LCD2_DATA1_PIN == 1) && (LCD2_DATA2_PIN == 2) && (LCD2_DATA3_PIN == 3) 
      && (LCD2_RS_PIN == 4 ) && (LCD2_RW_PIN == 5) && (LCD2_E_PIN == 6 ) )
    {
        /* configure all port bits as output (all LCD2 lines on same port) */
        DDR(LCD2_DATA0_PORT) |= 0x7F;
    }
    else if ( ( &LCD2_DATA0_PORT == &LCD2_DATA1_PORT) && ( &LCD2_DATA1_PORT == &LCD2_DATA2_PORT ) && ( &LCD2_DATA2_PORT == &LCD2_DATA3_PORT )
           && (LCD2_DATA0_PIN == 0 ) && (LCD2_DATA1_PIN == 1) && (LCD2_DATA2_PIN == 2) && (LCD2_DATA3_PIN == 3) )
    {
        /* configure all port bits as output (all LCD2 data lines on same port, but control lines on different ports) */
        DDR(LCD2_DATA0_PORT) |= 0x0F;
        DDR(LCD2_RS_PORT)    |= _BV(LCD2_RS_PIN);
        DDR(LCD2_RW_PORT)    |= _BV(LCD2_RW_PIN);
        DDR(LCD2_E_PORT)     |= _BV(LCD2_E_PIN);
    }
    else
    {
        /* configure all port bits as output (LCD2 data and control lines on different ports */
        DDR(LCD2_RS_PORT)    |= _BV(LCD2_RS_PIN);
        DDR(LCD2_RW_PORT)    |= _BV(LCD2_RW_PIN);
        DDR(LCD2_E_PORT)     |= _BV(LCD2_E_PIN);
        DDR(LCD2_DATA0_PORT) |= _BV(LCD2_DATA0_PIN);
        DDR(LCD2_DATA1_PORT) |= _BV(LCD2_DATA1_PIN);
        DDR(LCD2_DATA2_PORT) |= _BV(LCD2_DATA2_PIN);
        DDR(LCD2_DATA3_PORT) |= _BV(LCD2_DATA3_PIN);
    }
    delay(16000);        /* wait 16ms or more after power-on       */
    
    /* initial write to lcd2 is 8bit */
    LCD2_DATA1_PORT |= _BV(LCD2_DATA1_PIN);  // _BV(LCD2_FUNCTION)>>4;
    LCD2_DATA0_PORT |= _BV(LCD2_DATA0_PIN);  // _BV(LCD2_FUNCTION_8BIT)>>4;
    lcd2_e_toggle();
    delay(4992);         /* delay, busy flag can't be checked here */
   
    /* repeat last command */ 
    lcd2_e_toggle();      
    delay(64);           /* delay, busy flag can't be checked here */
    
    /* repeat last command a third time */
    lcd2_e_toggle();      
    delay(64);           /* delay, busy flag can't be checked here */

    /* now configure for 4bit mode */
    LCD2_DATA0_PORT &= ~_BV(LCD2_DATA0_PIN);   // LCD2_FUNCTION_4BIT_1LINE>>4
    lcd2_e_toggle();
    delay(64);           /* some displays need this additional delay */
    
    /* from now the LCD2 only accepts 4 bit I/O, we can use lcd2_command() */    
#else
    /*
     * Initialize LCD2 to 8 bit memory mapped mode
     */
    
    /* enable external SRAM (memory mapped lcd2) and one wait state */        
    MCUCR = _BV(SRE) | _BV(SRW);

    /* reset LCD2 */
    delay(16000);                           /* wait 16ms after power-on     */
    lcd2_write(LCD2_FUNCTION_8BIT_1LINE,0);   /* function set: 8bit interface */                   
    delay(4992);                            /* wait 5ms                     */
    lcd2_write(LCD2_FUNCTION_8BIT_1LINE,0);   /* function set: 8bit interface */                 
    delay(64);                              /* wait 64us                    */
    lcd2_write(LCD2_FUNCTION_8BIT_1LINE,0);   /* function set: 8bit interface */                
    delay(64);                              /* wait 64us                    */
#endif

#if KS0073_4LINES_MODE
    /* Display with KS0073 controller requires special commands for enabling 4 line mode */
	lcd2_command(KS0073_EXTENDED_FUNCTION_REGISTER_ON);
	lcd2_command(KS0073_4LINES_MODE);
	lcd2_command(KS0073_EXTENDED_FUNCTION_REGISTER_OFF);
#else
    lcd2_command(LCD2_FUNCTION_DEFAULT);      /* function set: display lines  */
#endif
    lcd2_command(LCD2_DISP_OFF);              /* display off                  */
    lcd2_clrscr();                           /* display clear                */ 
    lcd2_command(LCD2_MODE_DEFAULT);          /* set entry mode               */
    lcd2_command(dispAttr);                  /* display/cursor control       */

}/* lcd2_init */
